from importlib.metadata import version

from .factory import (
    assignment,
    cid,
    cohort,
    country,
    course,
    email,
    feedback,
    gender,
    mark,
    name,
    student,
    title,
    username,
)

__version__ = version("fakeitmakeit")

__all__ = [
    "cid",
    "gender",
    "title",
    "course",
    "country",
    "username",
    "email",
    "name",
    "mark",
    "feedback",
    "student",
    "cohort",
    "assignment",
]



import numpy as np
import pandas as pd
import pytest

import fakeitmakeit as fm


@pytest.fixture(scope="function")
def valid_assignment():
    return pd.Series(
        data=[23.0, 56.8, 72.5],
        index=pd.Index(["abc123", "def4561", "g789"], name="username"),
    )


@pytest.fixture(scope="function")
def invalid_assignment():
    return pd.Series(
        data=[23.0, -56.8, 72.5],
        index=pd.Index(["abc123", "def4561", "g789"], name="username"),
    )


@pytest.fixture(scope="function")
def valid_cohort():
    cohort = pd.DataFrame(
        {
            "cid": ["01079210", "02747269", "02605421"],
            "username": ["jsg8052", "tf97", "mk4717"],
            "github": ["acse-jsg8052", "edsml-tf97", "edsml-mk4717"],
            "course": ["acse", "edsml", "edsml"],
            "title": ["Miss", "Miss", "Mr"],
            "first_name": ["Jie", "Tracy", "Mahika"],
            "last_name": ["Gong", "Fry", "Kapur"],
            "gender": ["female", "female", "male"],
            "email": [
                "james13@imperial.ac.uk",
                "tchurch@imperial.ac.uk",
                "angela05@imperial.ac.uk",
            ],
            "tutor": ["Anne Spencer", "Theresa Jones", "Joseph Wells"],
            "fee_status": ["overseas", "home", "overseas"],
            "nationality": ["China", "United Kingdom", "India"],
            "enrollment_status": ["enrolled", "enrolled", "enrolled"],
            "personal_email": [
                "nathanolson@lindsey.biz",
                "ruizsally@smith.com",
                "willissandy@cline.com",
            ],
        }
    )
    return cohort.set_index("username")


@pytest.fixture(scope="module")
def invalid_cohort():
    # Duplicate username
    cohort = pd.DataFrame(
        {
            "cid": ["01079210", "02747269", "02605421"],
            "username": ["jsg8052", "tf97", "tf97"],
            "github": ["acse-jsg8052", "edsml-tf97", "edsml-mk4717"],
            "course": ["acse", "edsml", "edsml"],
            "title": ["Miss", "Miss", "Mr"],
            "first_name": ["Jie", "Tracy", "Mahika"],
            "last_name": ["Gong", "Fry", "Kapur"],
            "gender": ["female", "female", "male"],
            "email": [
                "james13@imperial.ac.uk",
                "tchurch@imperial.ac.uk",
                "angela05@imperial.ac.uk",
            ],
            "tutor": ["Anne Spencer", "Theresa Jones", "Joseph Wells"],
            "fee_status": ["overseas", "home", "overseas"],
            "nationality": ["China", "United Kingdom", "India"],
            "enrollment_status": ["enrolled", "enrolled", "enrolled"],
            "personal_email": [
                "nathanolson@lindsey.biz",
                "ruizsally@smith.com",
                "willissandy@cline.com",
            ],
        }
    )
    return cohort.set_index("username")


class TestEmail:
    def test_valid(self):
        # Check that valid emails are valid.
        assert fm.isvalid.email("username@domain.com")
        assert fm.isvalid.email("user.name@domain.com")
        assert fm.isvalid.email("user_name@domain.ac.uk")

    def test_invalid(self):
        # Check that invalid emails are invalid.
        assert not fm.isvalid.email("username@domain")
        assert not fm.isvalid.email("user_name@domain.")
        assert not fm.isvalid.email("user name@domain.com")
        assert not fm.isvalid.email("username@domain.com ")
        assert not fm.isvalid.email(" username@domain.com ")


class TestUsername:
    def test_valid(self):
        # Test cases with valid usernames
        assert fm.isvalid.username("s12")  # we allow one letter in username
        assert fm.isvalid.username("ab12")
        assert fm.isvalid.username("ab123")
        assert fm.isvalid.username("abc1234")
        assert fm.isvalid.username("xy9999")
        assert fm.isvalid.username("x99999")  # we allow 5 digits in username

    def test_invalid_start_digit(self):
        # Test case with a username starting with a digit
        assert not fm.isvalid.username("1ab123")

    def test_invalid_wrong_first_digit(self):
        # Test case with a username having '0' as the first digit after letters
        assert not fm.isvalid.username("ab0234")

    def test_invalid_too_few_digits(self):
        # Test case with too few digits
        assert not fm.isvalid.username("abc1")

    def test_invalid_too_many_digits(self):
        # Test case with too many digits
        assert not fm.isvalid.username("ab123456")

    def test_invalid_uppercase_letters(self):
        # Test case with uppercase letters
        assert not fm.isvalid.username("aBc123")

    def test_invalid_special_characters(self):
        # Test case with special characters
        assert not fm.isvalid.username("ab#123")

    def test_invalid_too_many_letters(self):
        # Test case with too many letters
        assert not fm.isvalid.username("abcd1234")

    def test_invalid_empty_string(self):
        # Test case with an empty string
        assert not fm.isvalid.username("")

    def test_invalid_whitespace(self):
        # Test case with whitespace
        assert not fm.isvalid.username("ab 1234")

    def test_invalid_trailing_whitespace(self):
        # Test case with whitespace
        assert not fm.isvalid.username("ab1234 ")
        assert not fm.isvalid.username(" ab1234")


class TestCID:
    def test_valid(self):
        # Test valid CIDs that meet all the criteria
        assert fm.isvalid.cid("01234567")
        assert fm.isvalid.cid("02123456")
        assert fm.isvalid.cid("05123456")
        assert fm.isvalid.cid("00123456")  # we allow two leading zeros

    def test_invalid_length(self):
        # Test CIDs with incorrect length (should be exactly 8 digits)
        assert not fm.isvalid.cid("0123456")  # Too short
        assert not fm.isvalid.cid("012345678")  # Too long

    def test_invalid_starting_digit(self):
        # Test a CID that does not start with a digit
        assert not fm.isvalid.cid("51234567")

    def test_invalid_contains_non_digit(self):
        # Test CIDs that contain non-digit characters
        assert not fm.isvalid.cid("01234a67")

    def test_invalid_empty_string(self):
        # Test an empty string (should be invalid)
        assert not fm.isvalid.cid("")

    def test_invalid_special_characters(self):
        # Test CIDs that contain special characters
        assert not fm.isvalid.cid("0123-567")

    def test_invalid_trailing_whitespace(self):
        # Test a CID with trailing whitespace
        assert not fm.isvalid.cid("01234567 ")
        assert not fm.isvalid.cid(" 01234567")


class TestName:
    def test_valid_single_word(self):
        # Test valid single word names (each word must start with a capital letter)
        assert fm.isvalid.name("John")
        assert fm.isvalid.name("Mary")

    def test_valid_multiple_words(self):
        # Test valid multiple word names (names with more than one word, each starting
        # with a capital letter)
        assert fm.isvalid.name("John Smith")
        assert fm.isvalid.name("Jean-Luc Picard")
        assert fm.isvalid.name("Jean-Luc (Peter) Picard")

    def test_valid_hyphenated_names(self):
        # Test valid hyphenated names (names with hyphens, each part starting with a
        # capital letter)
        assert fm.isvalid.name("Mary-Ann")
        assert fm.isvalid.name("Jean-Paul")

    def test_valid_accented_characters(self):
        # Test valid names with accented characters
        assert fm.isvalid.name("Željko Petrović")
        assert fm.isvalid.name("Jürgen Müller")
        assert fm.isvalid.name("Émilie Durand")
        assert fm.isvalid.name("José García")
        assert fm.isvalid.name("Łukasz Kowalski")
        assert fm.isvalid.name("Márquez López")
        assert fm.isvalid.name("François Dupont")
        assert fm.isvalid.name("Anaïs Fournier")
        assert fm.isvalid.name("Mesut Özil")
        assert fm.isvalid.name("Šimon Novak")

    def test_invalid_capitalization(self):
        # Test names with invalid capitalization (names must start each word with a
        # capital letter)
        assert not fm.isvalid.name("john")  # Not capitalized
        assert not fm.isvalid.name("JOHN")  # All uppercase

    def test_invalid_contains_digits_or_characters(self):
        # Test names containing digits or special characters (names should only contain
        # letters, spaces, or hyphens)
        assert not fm.isvalid.name("John123")  # Contains digits
        assert not fm.isvalid.name("John@Doe")  # Contains special characters
        assert not fm.isvalid.name("John_Doe")  # Contains special characters

    def test_invalid_use_of_hyphen_or_space(self):
        # Test names with incorrect use of hyphen or space (hyphens should replace
        # spaces, not be used alongside them)
        assert not fm.isvalid.name("John - Smith")  # Hyphen not used correctly
        assert not fm.isvalid.name("John  Smith")  # Double space

    def test_invalid_empty_or_whitespace(self):
        # Test empty string or names with leading/trailing whitespace (names should not
        # have extra whitespace)
        assert not fm.isvalid.name("")  # Empty string
        assert not fm.isvalid.name(" John")  # Leading space
        assert not fm.isvalid.name("John ")  # Trailing space

    def test_parenthesis(self):
        # Test names with parenthesis.
        assert fm.isvalid.name("John (Daniel) Smith")
        assert fm.isvalid.name("John (Daniel)")


class TestTitle:
    @pytest.mark.parametrize(
        "title, expected",
        [
            ("Mr", True),
            ("Ms", True),
            ("Mrs", True),
            ("Mx", True),
            ("Miss", True),
            ("Dr", True),
            ("mr", False),  # Not correctly capitalized
            ("Doctor", False),  # Not an exact match
            ("Lord", False),  # Not in the list
            ("Mx ", False),  # Trailing whitespace
            ("", False),  # Empty string
        ],
    )
    def test_valid(self, title, expected):
        assert fm.isvalid.title(title) == expected


class TestCourse:
    @pytest.mark.parametrize(
        "course, expected",
        [
            ("acse", True),
            ("edsml", True),
            ("gems", True),
            ("ACSE", False),  # Not lowercase
            ("Edsml", False),  # Not lowercase
            ("gEmS", False),  # Not lowercase
            ("math", False),  # Not in the list
            ("acse ", False),  # Trailing whitespace
            ("", False),  # Empty string
        ],
    )
    def test_valid(self, course, expected):
        assert fm.isvalid.course(course) == expected


class TestGender:
    @pytest.mark.parametrize(
        "gender, expected",
        [
            ("male", True),
            ("female", True),
            ("nonbinary", True),
            ("Male", False),  # Not lowercase
            ("FEMALE", False),  # Not lowercase
            ("NonBinary", False),  # Not lowercase
            ("man", False),  # Not in the list
            ("woman", False),  # Not in the list
            ("male ", False),  # Trailing whitespace
            ("", False),  # Empty string
        ],
    )
    def test_valid(self, gender, expected):
        assert fm.isvalid.gender(gender) == expected


class TestFeeStatus:
    @pytest.mark.parametrize(
        "fee_status, expected",
        [
            ("home", True),  # Valid fee status
            ("overseas", True),  # Valid fee status
            ("home - elq", True),  # Valid fee status
            ("Home", False),  # Not lowercase
            ("OVERSEAS", False),  # Not lowercase
            ("local", False),  # Not in the list
            ("international", False),  # Not in the list
            ("overseas ", False),  # Trailing whitespace
            ("", False),  # Empty string
        ],
    )
    def test_valid(self, fee_status, expected):
        assert fm.isvalid.fee_status(fee_status) == expected


class TestCountry:
    @pytest.mark.parametrize(
        "country, expected",
        [
            ("United Kingdom", True),  # Valid country
            ("United States", True),  # Valid country
            ("Croatia", True),  # Valid country
            ("United Kingdom ", False),  # Trailing whitespace
            ("", False),  # Empty string
            ("UK", False),  # Not in the list
            ("England", False),  # Not in the list
        ],
    )
    def test_valid(self, country, expected):
        assert fm.isvalid.country(country) == expected


class TestMark:
    @pytest.mark.parametrize(
        "mark, expected",
        [
            (0, True),  # Lower bound
            (100, True),  # Upper bound
            (50.0, True),  # Middle value
            (65.12, True),  # Middle value
            (101, False),  # Above upper bound
            (np.nan, True),  # Missing value
            (-1, False),  # Below lower bound
            ("50", False),  # Not a number
        ],
    )
    def test_valid(self, mark, expected):
        assert fm.isvalid.mark(mark) == expected


class TestAssignment:
    def test_valid(self, valid_assignment):
        # Check that valid assignments are valid
        assert fm.isvalid.assignment(valid_assignment)

    def test_with_valid_usernames(self, valid_assignment):
        assert fm.isvalid.assignment(
            valid_assignment, valid_usernames=["abc123", "def4561", "g789"]
        )

    def test_with_wrong_valid_usernames1(self, valid_assignment):
        # One username is missing.
        assert not fm.isvalid.assignment(
            valid_assignment, valid_usernames=["def4561", "g789"]
        )

    def test_with_wrong_valid_usernames2(self, valid_assignment):
        # One username is missing.
        with pytest.raises(ValueError):
            fm.isvalid.assignment(
                valid_assignment, valid_usernames=["def4561", "wrong_username"]
            )

    def test_wrong_index(self, valid_assignment):
        valid_assignment.index = ["1", "2", "3"]
        assert not fm.isvalid.assignment(valid_assignment)

    def test_index_name(self, valid_assignment):
        valid_assignment.index.name = "wrong_name"
        assert not fm.isvalid.assignment(valid_assignment)

    def test_index_not_unique(self, valid_assignment):
        valid_assignment.index = ["abc123", "abc123", "gr789"]
        valid_assignment.index.name = "username"
        assert not fm.isvalid.assignment(valid_assignment)

    def test_invalid(self, invalid_assignment):
        # Check that invalid assignments are invalid
        assert not fm.isvalid.assignment(invalid_assignment)


class TestCohort:
    def test_valid(self, valid_cohort):
        # Check that valid cohorts are valid
        assert fm.isvalid.cohort(valid_cohort)

    def test_invalid(self, invalid_cohort):
        assert not fm.isvalid.cohort(invalid_cohort)

    def test_wrong_index(self, valid_cohort):
        valid_cohort.index = ["1", "2", "3"]
        assert not fm.isvalid.cohort(valid_cohort)

    def test_wrong_index_name(self, valid_cohort):
        valid_cohort.index.name = "wrong_name"
        assert not fm.isvalid.cohort(valid_cohort)

    def test_wrong_data(self, valid_cohort):
        valid_cohort.loc["tf97", "first_name"] = "WRONG NAME"
        assert not fm.isvalid.cohort(valid_cohort)

    def test_github(self, valid_cohort):
        course = valid_cohort.github.str.extract(r"(^.*?)\-", expand=False)
        username = valid_cohort.github.str.extract(r"\-(.*?$)", expand=False)
        assert course.map(fm.isvalid.course).all()
        assert username.map(fm.isvalid.username).all()



import random
from dataclasses import dataclass

import faker
import pycountry

# Distributions.
GENDERS = {"male": 0.49, "female": 0.5, "nonbinary": 0.01}
TITLES = ["Mr", "Ms", "Mrs", "Miss", "Mx"]
COURSES = {"acse": 0.4, "edsml": 0.4, "gems": 0.2}
COUNTRIES = {country.name: 1 for country in pycountry.countries}


# Country-locale mapping.
# Exposed as mapping instead of a function for performance reasons.
COUNTRY_LOCALE = {}
for country in COUNTRIES:
    for locale in faker.config.AVAILABLE_LOCALES:
        if pycountry.countries.get(name=country).alpha_2 in locale:
            COUNTRY_LOCALE[country] = locale
            break


def discrete_draw(distribution):
    """Draw a value from a discrete distribution.

    Parameters
    ----------
    distribution: dict

        Dictionary with keys being possible outputs and values their probabilities.

    Returns
    -------
    str

        Randomly selected value.

    Examples
    --------
    >>> import fakeitmakeit as fm
    >>> fm.util.discrete_draw({"a": 0.5, "b": 0.5})  # doctest: +SKIP
    'a'

    """
    values = list(distribution.keys())
    probabilities = list(distribution.values())
    return random.choices(values, weights=probabilities, k=1)[0]


@dataclass
class Student:
    """A dataclass to be populated in student function."""

    cid: str
    gender: str
    nationality: str
    first_name: str
    last_name: str
    title: str
    course: str
    username: str
    email: str
    personal_email: str
    github: str
    fee_status: str
    enrollment_status: str
    tutor: str


@dataclass
class CohortBias:
    """A dataclass for grouping cohort-specific distributions."""

    gender: dict
    course: dict
    country_bias: dict


cohort_bias = CohortBias(
    gender={"male": 0.65, "female": 0.34, "nonbinary": 0.01},
    course={"acse": 0.4, "edsml": 0.4, "gems": 0.2},
    country_bias={
        "China": 1800,
        "United Kingdom": 350,
        "India": 150,
        "United States": 100,
        "Germany": 100,
        "France": 100,
        "Hong Kong": 100,
        "Spain": 100,
        "Italy": 100,
        "Netherlands": 80,
        "Canada": 80,
    },
)



import collections
import re

import numpy as np
import pandas as pd
import pytest

import fakeitmakeit as fm

UINT_RE = re.compile(r"^[0-9]+$")  # unsigned integer regex


@pytest.fixture(scope="module")
def cohort():
    return fm.cohort(n=100)


@pytest.fixture(scope="module")
def assignment(cohort):
    return fm.assignment(usernames=cohort.index, mean=65, std=10, pfail=0.1, pnan=0.1)


class TestCID:
    def test_type(self):
        # Check that the CID is a string.
        assert isinstance(fm.cid(), str)

    def test_length(self):
        # Check that the CID is 8 digits long.
        assert len(fm.cid()) == 8

    def test_all_digits(self):
        # Check that the CID is all digits.
        assert UINT_RE.search(fm.cid())

    def test_first_digit(self):
        # Check that the first digit is always 0.
        assert int(fm.cid()[0]) == 0

    def test_second_digit(self):
        # Check that the second digit is always 1 or 2.
        assert int(fm.cid()[1]) in [1, 2]

    def test_remaining_digits(self):
        # Check that the remaining 6 digits are always between 0 and 9.
        assert UINT_RE.search(fm.cid()[2:])

    def test_all_digits_present(self):
        # Check that all digits (0-9) can be present in the CID.
        assert len(set("".join(fm.cid() for _ in range(50)))) == 10

    def test_convertible(self):
        # Check that the CID is convertible to an integer.
        assert isinstance(int(fm.cid()), int)

    def test_isvalid(self):
        # Check that the CID is valid.
        assert fm.isvalid.cid(fm.cid())


class TestGender:
    def test_type(self):
        # Check that gender is a string.
        assert isinstance(fm.gender(), str)

    def test_values(self):
        # Check the output is one of the expected ones.
        assert fm.gender() in fm.util.GENDERS

    def test_probability_certain_outcome(self):
        # Check that the probabilities are respected.
        assert fm.gender(distribution={"outcome1": 1, "outcome2": 0}) == "outcome1"

    def test_probabilities_uncertain_outcome(self):
        # Check that the probabilities are respected.
        distribution = {"male": 0.2, "female": 0.75, "nonbinary": 0.05}
        counts = collections.Counter(
            fm.gender(distribution=distribution) for _ in range(1000)
        )
        assert counts["nonbinary"] < counts["male"] < counts["female"]

    def test_isvalid(self):
        # Check that gender is valid.
        assert fm.isvalid.gender(fm.gender())


class TestTitle:
    def test_type(self):
        # Check that title is a string.
        assert isinstance(fm.title(), str)

    def test_startswith_upper(self):
        # Check that the title starts with an uppercase M.
        assert fm.title().startswith("M")

    def test_no_gender(self):
        # Check that the title is one of the expected ones.
        assert fm.title() in fm.util.TITLES

    def test_male(self):
        # Check that the title is one of the expected ones for a male.
        assert fm.title(genderval="male") == "Mr"

    def test_female(self):
        # Check that the title is one of the expected ones for a female.
        assert fm.title(genderval="female") in {"Ms", "Mrs", "Miss"}

    def test_nonbinary(self):
        # Check that the title is one of the expected ones for a nonbinary.
        assert fm.title(genderval="nonbinary") == "Mx"

    def test_wrong_gender(self):
        # Check the exception is raised.
        with pytest.raises(ValueError):
            fm.title(genderval="wrong_gender")

    def test_isvalid(self):
        # Check that title is valid.
        assert fm.isvalid.title(fm.title())


class TestCourse:
    def test_type(self):
        # Check that course is a string.
        assert isinstance(fm.course(), str)

    def test_values(self):
        # Check the output is one of the expected ones.
        assert fm.course() in fm.util.COURSES

    def test_probability_certain_outcome(self):
        # Check that the probabilities are respected.
        assert fm.course(distribution={"course1": 1, "course2": 0}) == "course1"

    def test_probabilities_uncertain_outcome(self):
        # Check that the probabilities are respected.
        distribution = {"c1": 0.2, "c2": 0.75, "c3": 0.05}
        counts = collections.Counter(
            fm.course(distribution=distribution) for _ in range(1000)
        )
        assert counts["c3"] < counts["c1"] < counts["c2"]

    def test_isvalid(self):
        # Check that course is valid.
        assert fm.isvalid.course(fm.course())


class TestCountry:
    def test_type(self):
        # Check that country is a string.
        assert isinstance(fm.country(), str)

    def test_defaults(self):
        # Check that the country is one of the expected ones.
        assert fm.country() in fm.util.COUNTRIES

    def test_probability_certain_outcome(self):
        # Check that the probabilities are respected.
        assert fm.country(distribution={"country1": 1, "country2": 0}) == "country1"

    def test_probabilities_uncertain_outcome(self):
        # Check that the probabilities are respected.
        distribution = {"c1": 0.2, "c2": 0.75, "c3": 0.05}
        counts = collections.Counter(
            fm.country(distribution=distribution) for _ in range(1000)
        )
        assert counts["c3"] < counts["c1"] < counts["c2"]

    def test_bias(self):
        # Check that the bias is respected.
        bias = {"China": 1000, "United Kingdom": 100}
        counts = collections.Counter(fm.country(bias=bias) for _ in range(100))
        assert counts["China"] > counts["United Kingdom"] > counts["Syria"]

    def test_isvalid(self):
        # Check that country is valid.
        assert fm.isvalid.country(fm.country())


class TestUsername:
    def test_type(self):
        # Check that username is a string.
        assert isinstance(fm.username(), str)

    def test_length(self):
        # Check that the username is between 4 and 7 characters long.
        assert 4 <= len(fm.username()) <= 7

    def test_lower(self):
        # Check that the username is lowercase.
        assert fm.username().islower()

    def test_no_space(self):
        # Check that the username does not contain spaces.
        assert " " not in fm.username()

    def test_first_letter(self):
        # Check that the first letter is the first letter of the first name.
        assert fm.username(nameval="John Smith").startswith("j")

    def test_last_letter(self):
        # Check that the last letter is the first letter of the last name.
        letters = [i for i in fm.username(nameval="John Doe") if i.isalpha()]
        assert letters[-1] == "d"

    def test_num_digits(self):
        # Check that the username contains 2 to 4 digits.
        assert 2 <= sum(i.isdigit() for i in fm.username()) <= 4

    def test_num_letters(self):
        # Check that the username contains 2 or 3 letters.
        assert 2 <= sum(i.isalpha() for i in fm.username()) <= 3

    def test_isvalid(self):
        # Check that username is valid.
        assert fm.isvalid.username(fm.username())


class TestEmail:
    def test_type(self):
        # Check that email is a string.
        assert isinstance(fm.email(), str)

    def test_at(self):
        # Check that the email contains an @.
        assert "@" in fm.email()

    def test_dot(self):
        # Check that the email contains a dot.
        assert "." in fm.email()

    def test_domain(self):
        # Check that the email contains domainval if specified.
        assert fm.email(domainval="gmail.com").endswith("@gmail.com")

    def test_valid(self):
        # Check that the email is valid.
        assert fm.isvalid.email(fm.email())


class TestName:
    def test_type(self):
        # Check that name is a string.
        assert isinstance(fm.name(), str)

    def test_space(self):
        # Check that the name contains a space.
        assert " " in fm.name()

    def test_china(self):
        # Check that Chinese names are generated as expected.
        for i in range(300):
            first_name, *_ = fm.name(countryval="China").split()
            if first_name in {
                "Jang",
                "Jing",
                "Wei",
                "Fang",
                "Lei",
                "Tao",
                "Qiang",
                "Ming",
                "Chao",
                "Li",
            }:
                assert True
                break
        else:
            assert False

    def test_gender(self):
        # Check expected female names are in the output.
        for i in range(300):
            first_name, *_ = fm.name(
                genderval="female", countryval="United Kingdom"
            ).split()
            if first_name in {
                "Ellie",
                "Jill",
                "Irene",
                "Jean",
                "Megan",
                "Fiona",
                "Sylvia",
                "Claire",
                "Kim",
                "Lydia",
                "Jane",
                "Karen",
                "Amy",
                "Paula",
            }:
                assert True
                break
        else:
            assert False

    def test_isvalid(self):
        # Check that name is valid.
        assert fm.isvalid.name(fm.name())


class TestMark:
    def test_type(self):
        # Check that mark is a float.
        # np.nan is an instance of float as well.
        assert isinstance(fm.mark(), float)

    def test_range(self):
        # Check that the mark is between 0 and 100.
        assert 0 <= fm.mark(pnan=0) <= 100

    def test_pfail_1(self):
        # Check pfail is respected.
        assert fm.mark(pfail=1, pnan=0) == 0.0

    def test_pfail_0_5(self):
        # Check pfail is respected.
        marks = np.array([fm.mark(pfail=0.5, pnan=0) for _ in range(100)])
        assert 35 <= (marks == 0).sum() <= 65

    def test_pfail_0(self):
        # Check pfail is respected.
        marks = np.array([fm.mark(pfail=0, pnan=0) for _ in range(100)])
        assert (marks == 0).sum() == 0

    def test_pnan_1(self):
        # Check pfail is respected.
        assert np.isnan(fm.mark(pfail=0, pnan=1))

    def test_pnan_0_5(self):
        # Check pfail is respected.
        marks = np.array([fm.mark(pfail=0, pnan=0.5) for _ in range(100)])
        assert 35 <= np.isnan(marks).sum() <= 65

    def test_pnan_0(self):
        # Check pfail is respected.
        marks = np.array([fm.mark(pfail=0, pnan=0) for _ in range(100)])
        assert np.isnan(marks).sum() == 0

    def test_mean(self):
        # Check that the mean is as expected.
        marks = np.array(
            [fm.mark(mean=65, std=10, pfail=0, pnan=0) for _ in range(100)]
        )
        assert 60 <= marks.mean() <= 70

    def test_std(self):
        # Check that the standard deviation is as expected.
        marks = np.array(
            [fm.mark(mean=65, std=10, pfail=0, pnan=0) for _ in range(100)]
        )
        assert 8 <= marks.std() <= 12

    def test_isvalid(self):
        # Check that mark is valid.
        assert fm.isvalid.mark(fm.mark())


class TestFeedback:
    def test_type(self):
        # Check that feedback is a string.
        assert isinstance(fm.feedback(), str)


class TestStudent:
    def test_type(self):
        # Check that the output is a Student dataclass.
        assert isinstance(fm.student(), fm.util.Student)

    def test_cid(self):
        # Check that the CID is a string.
        assert hasattr(fm.student(), "cid")

    def test_repr(self):
        # Check that the repr string makes sense.
        assert "Student" in repr(fm.student())


class TestCohort:
    def test_type(self, cohort):
        # Check that the output is a DataFrame.
        assert isinstance(cohort, pd.DataFrame)

    def test_course(self, cohort):
        # Check that the output is the right length.
        assert len(cohort["course"].unique()) == 3

    def test_tutor(self, cohort):
        # Check that the number of tutors is bounded.
        assert cohort["tutor"].map(fm.isvalid.name).all()

    def test_nationality(self, cohort):
        # Check that nationalities are as expected.
        counts = collections.Counter(cohort["nationality"])

        assert set(counts.keys()) <= set(fm.util.COUNTRIES.keys())
        assert counts["China"] > 40
        assert counts["United Kingdom"] > 4

    def test_username(self, cohort):
        # Check that usernames are as expected.
        assert cohort.index.name == "username"
        assert cohort.index.is_unique
        assert cohort.index.map(fm.isvalid.username).all()

    def test_cid(self, cohort):
        # Check that CIDs are as expected.
        assert cohort["cid"].map(fm.isvalid.cid).all()

    def test_email(self, cohort):
        # Check that emails are as expected.
        assert cohort["email"].map(fm.isvalid.email).all()

    def test_personal_email(self, cohort):
        # Check that emails are as expected.
        assert cohort["personal_email"].map(fm.isvalid.email).all()

    def test_title(self, cohort):
        # Check that titles are as expected.
        assert cohort["title"].map(fm.isvalid.title).all()

    def test_first_name(self, cohort):
        # Check that names are as expected.
        assert cohort["first_name"].map(fm.isvalid.name).all()

    def test_last_name(self, cohort):
        # Check that names are as expected.
        assert cohort["last_name"].map(fm.isvalid.name).all()

    def test_gender(self, cohort):
        # Check that genders are as expected.
        assert cohort["gender"].map(fm.isvalid.gender).all()

    def test_fee_status(self, cohort):
        # Check that genders are as expected.
        assert cohort["fee_status"].map(fm.isvalid.fee_status).all()

    def test_isvalid(self, cohort):
        # Check that the output is a DataFrame.
        assert fm.isvalid.cohort(cohort)


class TestAssignment:
    def test_type(self, assignment):
        # Check that the output is a DataFrame.
        assert isinstance(assignment, pd.Series)
        assert assignment.dtype == np.float64

    def test_index(self, assignment):
        # Check that usernames are as expected.
        assert assignment.index.is_unique
        assert assignment.index.name == "username"
        assert assignment.index.map(fm.isvalid.username).all()

    def test_mark(self, assignment):
        # Check that the marks are as expected.
        assert assignment.apply(fm.isvalid.mark).all()

    def test_wrong_username(self):
        # Check that the exception is raised.
        with pytest.raises(ValueError):
            fm.assignment(usernames=["wrong_username", "abc123"])

    def test_isvalid(self, cohort):
        assignment = fm.assignment(usernames=cohort.index, pnan=0.1)
        assert fm.isvalid.assignment(assignment, valid_usernames=cohort.index)

    def test_nan(self, assignment):
        # Ensure that there are np.nan values in the assignment.
        assert assignment.isna().sum() > 0



import numbers
import re

import fakeitmakeit as fm


def test_version():
    # Check that the version is a string.
    assert isinstance(fm.__version__, str)
    assert re.fullmatch(r"20[0-9]{2}\.[0-9]{1,2}\.[0-9]{1,2}", fm.__version__)


class TestGENDERS:
    def test_type(self):
        # Check that GENDERS is a dictionary.
        assert isinstance(fm.util.GENDERS, dict)

    def test_length(self):
        # Check there are 3 genders.
        assert len(fm.util.GENDERS) == 3

    def test_keys(self):
        # Check that the keys are genders.
        assert "male" in fm.util.GENDERS

    def test_values(self):
        # Check that the values are all numbers.
        assert all(
            isinstance(value, numbers.Real) for value in fm.util.GENDERS.values()
        )


class TestTITLES:
    def test_type(self):
        # Check that TITLES is a set.
        assert isinstance(fm.util.TITLES, list)

    def test_length(self):
        # Check that TITLES has 5 titles.
        assert len(fm.util.TITLES) == 5

    def test_elements(self):
        # Check correct elements are in the set.
        assert "Mr" in fm.util.TITLES

    def test_values(self):
        # Check that the values are all strings.
        assert all(isinstance(value, str) for value in fm.util.TITLES)


class TestCOURSES:
    def test_type(self):
        # Check that COURSES is a dictionary.
        assert isinstance(fm.util.COURSES, dict)

    def test_length(self):
        # Check that COURSES has 3 courses.
        assert len(fm.util.COURSES) == 3

    def test_keys(self):
        # Check that the keys are course names.
        assert {"acse", "edsml", "gems"} <= set(fm.util.COURSES.keys())

    def test_values(self):
        # Check that the values are all numbers.
        assert all(
            isinstance(value, numbers.Real) for value in fm.util.COURSES.values()
        )


class TestCOUNTRIES:
    def test_type(self):
        # Check that COUNTRIES is a dictionary.
        assert isinstance(fm.util.COUNTRIES, dict)

    def test_length(self):
        # Check that COUNTRIES has sufficient items.
        assert len(fm.util.COUNTRIES) >= 245

    def test_keys(self):
        # Check that the keys are country names.
        assert {"Serbia", "Malta"} < set(fm.util.COUNTRIES.keys())

    def test_values(self):
        # Check that the values are all 1 (relative probabilities).
        assert all(
            isinstance(value, numbers.Real) for value in fm.util.COUNTRIES.values()
        )


class TestCOUNTRYLOCALE:
    def test_type(self):
        # Check that COUNTRY_LOCALE is a dictionary.
        assert isinstance(fm.util.COUNTRY_LOCALE, dict)

    def test_length(self):
        # Check that COUNTRY_LOCALE has sufficient items.
        assert len(fm.util.COUNTRY_LOCALE) >= 10

    def test_keys(self):
        # Check that the keys are country names.
        assert {"Argentina", "United States"} < set(fm.util.COUNTRY_LOCALE.keys())

    def test_values(self):
        # Check that the values contain an underscore.
        assert all("_" in value for value in fm.util.COUNTRY_LOCALE.values())

    def test_entries(self):
        # Check that the entries are valid locales.
        assert fm.util.COUNTRY_LOCALE["China"] == "zh_CN"
        assert fm.util.COUNTRY_LOCALE["United Kingdom"] == "en_GB"


class TestDiscreteDraw:
    def test_output(self):
        # Check that the output is one of the expected ones.
        assert fm.util.discrete_draw({"a": 0.5, "b": 0.5}) in {"a", "b"}

    def test_certain(self):
        # Check that the output is always the same if probability is 1.
        assert fm.util.discrete_draw({"a": 0, "b": 1}) == "b"



import logging
import numbers
import re

import numpy as np
import pandas as pd

import fakeitmakeit.util as fmu


def email(value):
    """Check if ``value`` is a valid email.

    Parameters
    ----------
    value: str

        Email.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.isvalid.email('nikola.tesla@gmail.com')
    True
    >>> fm.isvalid.email('nikola.tesla(at)gmail.com')
    False

    """
    email_re = r"[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}"
    return bool(re.fullmatch(email_re, value))


def username(value):
    """Check if ``value`` is a valid username.

    Parameters
    ----------
    value: str

        Username.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.isvalid.username('johndoe')
    False
    >>> fm.isvalid.username('abc1234')
    True

    """
    # We allow 1-3 lowercase letters followed by 1-5 numbers - first number is never 0.
    username_re = r"[a-z]{1,3}[1-9][0-9]{1,4}"
    return bool(re.fullmatch(username_re, value))


def cid(value):
    """Check if ``value`` is a valid CID.

    Parameters
    ----------
    value: str

        CID.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.isvalid.cid('01234567')
    True
    >>> fm.isvalid.cid('012345678')  # 9 digits
    False

    """
    cid_re = r"0[0-9]{7}"
    return bool(re.fullmatch(cid_re, value))


def name(value, allow_special_characters=True):
    """Check if ``value`` is a valid name.

    Parameters
    ----------
    value: str

        Name.

    allow_special_characters: bool, optional

        If ``True``, allow special characters in names, e.g. umlauts.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.isvalid.name('Albert Einstein')
    True
    >>> fm.isvalid.name('Albert')
    True
    >>> fm.isvalid.name('john')
    False
    >>> fm.isvalid.name('Nikola_Tesla')
    False

    """
    if allow_special_characters:
        word = r"[A-Z\u00C0-\u017F][a-z\u00C0-\u017F]"
    else:
        word = r"[A-Z][a-z]"

    name_re = rf"({word}*)([-\s](({word}*)|\({word}*\)))*"
    return bool(re.fullmatch(name_re, value))


def title(value):
    """Check if ``value`` is a valid title.

    Parameters
    ----------
    value: str

        Title.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.isvalid.title('Mr')
    True
    >>> fm.isvalid.title('mr')
    False
    >>> fm.isvalid.title('Herr')
    False

    """
    title_re = r"(Mr|Ms|Mrs|Mx|Miss|Dr)"
    return bool(re.fullmatch(title_re, value))


def course(value):
    """Check if ``value`` is a valid course.

    Parameters
    ----------
    value: str

        Course.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.isvalid.course('acse')
    True
    >>> fm.isvalid.course('a_very_cool_course')
    False

    """
    course_re = r"(acse|edsml|gems)"
    return bool(re.fullmatch(course_re, value))


def gender(value):
    """Check if ``value`` is a valid gender.

    Parameters
    ----------
    value: str

        Gender.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.isvalid.gender('female')
    True
    >>> fm.isvalid.gender('Female')
    False

    """
    gender_re = r"(male|female|nonbinary)"
    return bool(re.fullmatch(gender_re, value))


def fee_status(value):
    """Check if ``value`` is a valid fee status.

    Parameters
    ----------
    value: str

        Fee status.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.isvalid.fee_status('home')
    True
    >>> fm.isvalid.fee_status('overseas')
    True
    >>> fm.isvalid.fee_status('home1')
    False

    """
    fee_status_re = r"(home|overseas|(home - elq))"
    return bool(re.fullmatch(fee_status_re, value))


def country(value):
    """Check if ``value`` is a valid country.

    Parameters
    ----------
    value: str

        Country.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.isvalid.country('United Kingdom')
    True
    >>> fm.isvalid.country('Neverland')
    False

    """
    allowed_countries = set(fmu.COUNTRIES.keys()) | {
        "Taiwan",
        "Syria",
        "Columbia",
        "Turkey",
        "United States of America",
        "Russia",
    }
    return value in allowed_countries


def mark(value):
    """Check if ``value`` is a valid mark.

    For ``value`` to be valid, it must be:
    1. An instance of ``numbers.Real`` (float, int, or ``np.nan``).
    2. If the value is not ``np.nan``, it must be ``0 <= value <= 100``.

    Parameters
    ----------
    value: numbers.Real

        Mark.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    >>> import numpy as np
    ...
    >>> fm.isvalid.mark(100)
    True
    >>> fm.isvalid.mark(101)
    False
    >>> fm.isvalid.mark(-1)
    False
    >>> fm.isvalid.mark(np.nan)
    True
    >>> fm.isvalid.mark("100")
    False

    """
    if not isinstance(value, numbers.Real):
        logging.warning(f"Invalid type {type(value)=}.")
        return False
    elif not np.isnan(value) and not 0 <= value <= 100:
        logging.warning(f"{value=} is not in [0, 100] range.")
        return False
    else:
        return True


def assignment(value, valid_usernames=None):
    """Check if ``value`` is a valid assignment.

    For the assessment to be valid, the following conditions must be met:

    1. ``value`` must be a ``pd.Series``.
    2. Index values must be valid usernames.
    3. Index values must be unique.
    4. Index name must be "username".
    5. Data values must be valid marks (``float`` in [0, 100] range or np.nan).
    6. If ``valid_usernames`` is provided, all usernames in index must be in it.

    Parameters
    ----------
    value: pd.Series

        Assignment. Index values are usernames and data are numerical marks.

    valid_usernames: Iterable[str], optional

        Iterable of valid usernames.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``. Fail reason is logged with level
        ``warning``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    >>> import pandas as pd
    ...
    >>> value = pd.Series(
    ...     data = [50.1, 70],
    ...     index = pd.Index(["abc123", "sw4321"], name="username"),
    ...     name = "mark",
    ... )
    >>> fm.isvalid.assignment(value)
    True

    """
    # Check that value is a pd.Series.
    if not isinstance(value, pd.Series):
        logging.warning(f"Invalid type {type(value)=} - pd.Series expected.")
        return False

    # Check that indicies are valid usernames.
    valid = value.index.map(username)
    if not valid.all():
        logging.warning(f"Invalid usernames: {value.index[~valid].tolist()}")
        return False

    # Check if there are any duplicated usernames.
    duplicated = value.index.duplicated()
    if duplicated.any():
        logging.warning(f"Duplicated usernames: {value.index[duplicated].tolist()}.")
        return False

    # Check that index name is correct.
    if value.index.name != "username":
        logging.warning(
            f"Invalid index name {value.index.name=} - 'username' expected."
        )
        return False

    # Check data (marks).
    valid = value.map(mark)
    if not valid.all():
        logging.warning(f"Invalid marks: {value[~valid].tolist()}.")
        return False

    if valid_usernames is not None:
        valid_usernames = set(valid_usernames)
        invalid = [u for u in valid_usernames if not username(u)]
        if invalid:
            raise ValueError(f"Invalid usernames in valid_username: {invalid}.")

        valid = value.index.isin(valid_usernames)
        if not valid.all():
            logging.warning(f"Invalid usernames in index: {value.index[~valid]}.")
            return False

    return True


def cohort(value):
    """Check if ``value`` is a valid cohort.

    Parameters
    ----------
    value: pd.DataFrame

        Cohort.

    Returns
    -------
    bool

        ``True`` if valid, otherwise ``False``.

    """
    # Check that indicies are valid usernames.
    if not value.index.map(username).all():
        invalid = value.index[~value.index.map(username)]
        logging.warning(f"Invalid usernames: {invalid.tolist()}")
        return False

    # Check the index name.
    if value.index.name != "username":
        logging.warning(
            f"Invalid index name {value.index.name} - it must be 'username'."
        )
        return False

    # Check if there are any repeated usernames.
    if not value.index.is_unique:
        invalid = value.index[value.index.duplicated()]
        logging.warning(f"There are duplicate usernames: {invalid.tolist()}.")
        return False

    # Check other columns.
    for col in set(value.columns) - {
        "username",
        "github",
        "enrollment_status",
        "comment",
    }:
        if "name" in col or col == "tutor":
            data_type = "name"
        elif "email" in col:
            data_type = "email"
        elif col == "nationality":
            data_type = "country"
        else:
            data_type = col

        # print(col, data_type)

        validation_function = globals()[f"{data_type}"]
        if not value[col].map(validation_function).all():
            invalid = value[col][~value[col].map(validation_function)]
            logging.warning(f'Errors in column "{col}": {invalid.tolist()}.')
            return False
    return True



import random
import re
import string
from dataclasses import fields

import numpy as np
import pandas as pd
from faker import Faker

import fakeitmakeit.isvalid as fmiv
import fakeitmakeit.util as fmu


def cid():
    """Generate a random 8-digit CID.

    The first digit is always 0, whereas the second digit is 1 or 2. The remaining 6
    digits are randomly generated between 0 and 9.

    Returns
    -------
    str

        Randomly generated CID.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.cid()  # doctest: +SKIP
    "01234567"

    """
    # The first digit is always 0.
    number = "0"

    # The second digit is 1 or 2.
    number += str(random.choice([1, 2]))

    # Generate the remaining 6 digits randomly between 0 and 9.
    number += "".join(str(random.randint(0, 9)) for _ in range(6))

    return number


def gender(distribution=dict(fmu.GENDERS)):
    """Generate a random gender.

    Possible gender values and their relative probabilities are passed via
    ``distribution`` - a dictionary whose keys are possible outputs and values are
    relative probablilities. The values in ``distribution`` do not have to sum to 1
    because selections will be made according to the relative weights.

    Parameters
    ----------
    distribution: dict

        Keys are possible output values and values are relative probablilities. For
        instance, ``{"male": 0.5, "female": 0.5}``.

    Returns
    -------
    str

        Randomly generated gender.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.gender()  # doctest: +SKIP
    'male'

    """
    return fmu.discrete_draw(distribution)


def title(genderval=None):
    """Generate a random title.

    If gender is provided via ``genderval``, then the title is generated
    according to it. Otherwise, the gender and accordingly the title is
    generated randomly.

    For further details, refer to:
    - https://www.grammarly.com/blog/ms-mrs-miss-difference/
    - https://publishing.rcseng.ac.uk/doi/pdf/10.1308/rcsbull.2021.141

    Parameters
    ----------
    genderval: str

        Gender.

    Returns
    -------
    str

        Randomly generated title.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.title('male')
    'Mr'
    >>> fm.title('nonbinary')
    'Mx'
    >>> fm.title()  # doctest: +SKIP
    'Mx'

    """
    genderval = genderval or gender()
    if genderval == "male":
        return "Mr"
    elif genderval == "female":
        return random.choice(["Ms", "Mrs", "Miss"])
    elif genderval == "nonbinary":
        return "Mx"
    else:
        raise ValueError(f"Invalid gender: {genderval=}")


def course(distribution=dict(fmu.COURSES)):
    """Generate a random course.

    Possible courses and their relative probabilities are passed via ``distribution``.
    It is a dictionary whose keys are possible outputs and values are relative
    probablilities. The values in ``distribution`` do not have to sum to 1 because
    selections will be made according to the relative weights.

    Parameters
    ----------
    distribution: dict

        Keys are possible outputs and values are relative probablilities. For instance,
        ``{"course1": 0.5, "course2": 0.5}``.

    Returns
    -------
    str

        Randomly generated course.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.course()  # doctest: +SKIP
    'acse'

    """
    return fmu.discrete_draw(distribution)


def country(distribution=dict(fmu.COUNTRIES), bias=None):
    """Generate a random country.

    Distribution is passed via ``distribution``. It is a dictionary whose keys are
    possible outputs and values are relative probablilities. The values in
    ``distribution`` do not have to sum to 1 because selections will be made according
    to the relative weights.

    By default, all countries are equally likely and their relative probablility is 1.
    To modify the default probablilities, pass a dictionary of countries and their
    relative probabilities via ``bias``. Internally, ``distribution |= bias`` is
    calculated.

    Parameters
    ----------
    distribution: dict

        Keys are possible outputs and values are relative probablilities.

    Returns
    -------
    str

        Randomly generated country.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.country()  # doctest: +SKIP
    'United Kingdom'
    >>> fm.country(bias={'India': 0.5, 'United Kingdom': 0.2})  # doctest: +SKIP
    'India'
    >>> fm.country(distribution={}, bias={'France': 1})
    'France'

    """
    distribution |= bias or {}
    return fmu.discrete_draw(distribution)


def username(nameval=None):
    """Generate a random username.

    The username is the combination of 2-3 lowercase letters and a random number with 2
    to 4 digits. The first letter of the username is the first letter of the first name,
    whereas the the last letter of the username is the first letter of the last name.
    Letters are followed with a random integer between 2 and 4 digits. The first digit
    of the number is never zero.

    Parameters
    ----------
    nameval: str

        Name.

    Returns
    -------
    str

        Randomly generated username.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.username('John Smith')  # doctest: +SKIP
    'jws4122'

    """
    nameval = nameval or name()

    # Get the first letter of the first name.
    first_letter, *_ = nameval.casefold().split()[0]

    # Get the first letter of the last name.
    last_letter, *_ = nameval.casefold().split()[-1]

    # Randomly decide if the string will have 2 or 3 letters.
    if random.choice([2, 3]) == 3:
        # Generate a random middle lowercase letter (can be any lowercase letter).
        middle_letter = random.choice(string.ascii_lowercase)
        letters = first_letter + middle_letter + last_letter
    else:
        letters = first_letter + last_letter

    # Generate a random number between 2 and 4 digits where the first digit is not zero.
    num_digits = random.choice([2, 3, 4])
    numbers = str(random.randint(1, 9))  # First digit is never zero
    for _ in range(num_digits - 1):
        numbers += str(random.randint(0, 9))

    return letters + numbers


def email(domainval=None):
    """Generate a random email.

    If ``domainval`` is not provided, then an email with a random domain is generated.
    Otherwise, the email is generated with the provided domain.

    Parameters
    ----------
    domainval: str

        Domain.

    Returns
    -------
    str

        Randomly generated email.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.email()  # doctest: +SKIP
    'somerandomemail@domain.com'
    >>> fm.email(domainval='myuniversity.ac.uk')  # doctest: +SKIP
    'john.doe@myuniversity.ac.uk'

    """
    fake = Faker()
    domain = domainval or fake.domain_name()
    return f"{fake.user_name()}@{domain}"


def name(genderval=None, countryval=None):
    """Generate a random name.

    This function uses Faker to generate a random name. Depending on the ``countryval``
    and ``genderval`` parameters, the name is generated according to the locale and by
    calling the appropriate method from ``Faker``. Otherwise, the name with the default
    ``Faker`` locale is returned.

    Parameters
    ----------
    genderval: str

        Gender.

    countryval: str

        Country.

    Returns
    -------
    str

        Randomly generated name.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.name()  # doctest: +SKIP
    'John Smith'
    >>> fm.name(countryval='China')  # doctest: +SKIP
    'Zhang San'
    >>> fm.name(genderval='female', countryval='Germany')  # doctest: +SKIP
    'Anna Schmidt'

    """
    locale = fmu.COUNTRY_LOCALE.get(countryval, None)
    fake = Faker(locale)

    # Romanized is available only for some countries.
    if hasattr(fake, "romanized_name"):
        res = fake.romanized_name()
    else:
        # Depending on the gender, we call the appropriate method from Faker.
        method = f"name_{genderval}" if genderval is not None else "name"

        # Not all countries have names for different genders.
        try:
            res = getattr(fake, method)()
        except AttributeError:
            res = fake.name()

        # Remove suffixes and prefixes - Mr, PhD, words with dots and all caps.
        # This is not exhaustive and some names might still contain some of these.
        pattern = r"\b(?:[A-Z]+\b|PhD|Dr\(a\)|,|Dr|Mr|Mrs|Ms|Miss|\w*\.\w*)"
        res = re.sub(pattern, "", res).strip()

    if fmiv.name(res, allow_special_characters=False):
        return res
    else:
        # If the name is not valid, then we generate a new one with default faker until
        # it passes validation.
        while not fmiv.name(res, allow_special_characters=False):
            res = Faker().name()

        return res


def mark(mean=65.0, std=6.0, pfail=0.02, pnan=0.0):
    """Generate a random mark.

    A mark between 0 and 100 is generated from a normal distribution with the given
    ``mean`` and standard deviation ``std``. There is fail probability ``pfail`` that
    the mark will be 0. Similarly, the probability that mark will be ``np.nan`` is
    ``pnan. Resulting mark is rounded to two decimal places.

    Mark 0 can be interpreted as the student not submitting the assignment, whereas
    ``np.nan`` can be interpreted as the mark not being available, e.g. skipped exam due
    to mitigating circumstances.

    Parameters
    ----------
    mean: float

        Mean.

    std: float

        Standard deviation.

    pfail: float

        Probability that the mark will be 0.

    pnan: float

        Probability that the mark will be ``np.nan``.

    Returns
    -------
    float

        Randomly generated from [0, 100] range or ``np.nan``.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.mark(mean=65, std=10)  # doctest: +SKIP
    71
    >>> fm.mark(mean=65, std=10, pfail=1, pnan=0)
    0.0
    >>> fm.mark(mean=65, std=10, pfail=0, pnan=1)
    nan

    """
    rng = np.random.default_rng()

    if rng.random() < pfail:
        # Return 0 with a probability of pfail.
        return 0.0
    elif rng.random() < pnan:
        # Return np.nan with a probability of pfail.
        return np.nan
    else:
        # Generate a random number from a normal distribution with the given
        # mean and standard deviation.
        number = rng.normal(loc=mean, scale=std)
        # Clip the number to be be between 0 and 100.
        number = np.clip(number, 0, 100)
        # Round the number to two decimal places
        return round(number, 2)


def feedback():
    """Generate a random feedback.

    The feedback is generated by Faker and consists of one paragraph.

    Returns
    -------
    str

        Randomly generated feedback.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.feedback()  # doctest: +SKIP
    ...

    """
    return "\n\n".join(Faker().paragraphs(nb=1))


def student():
    """Generate a random student.

    Returns
    -------
    Student

        Student dataclass.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.student()
    Student(cid=...)

    """
    # (Intermediate) values required for other fields.
    genderval = gender(distribution=fmu.cohort_bias.gender)
    courseval = course(distribution=fmu.cohort_bias.course)
    countryval = country(bias=fmu.cohort_bias.country_bias)
    first_name, *_, last_name = name(genderval=genderval, countryval=countryval).split()
    usernameval = username(nameval=f"{first_name} {last_name}")

    return fmu.Student(
        cid=cid(),
        gender=genderval,
        course=courseval,
        nationality=countryval,
        first_name=first_name,
        last_name=last_name,
        title=title(genderval=genderval),
        username=usernameval,
        email=email(domainval="imperial.ac.uk"),
        personal_email=email(),
        github=f"{courseval}-{usernameval}",
        fee_status="home" if countryval == "United Kingdom" else "overseas",
        enrollment_status="enrolled",
        tutor=name(),
    )


def cohort(n):
    """Generate a cohort of students.

    Parameters
    ----------
    n: int

        Number of students.

    Returns
    -------
    pd.DataFrame

        A cohort dataframe.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.cohort(n=30)  # doctest: +SKIP
    ...

    """
    students = [student() for _ in range(n)]
    return (
        pd.DataFrame(
            {
                col.name: [getattr(student, col.name) for student in students]
                for col in fields(fmu.Student)
            }
        )
        .set_index("username", verify_integrity=True)
        .astype(
            {
                "course": "category",
                "gender": "category",
                "title": "category",
                "nationality": "category",
                "fee_status": "category",
                "enrollment_status": "category",
            }
        )
    )


def assignment(usernames, mean=65, std=6, pfail=0.02, pnan=0.0):
    """Generate an assignment.

    For each username, a mark is generated using the ``mark`` function.

    Parameters
    ----------
    usernames: Iterable[str]

        Iterable of usernames.

    mean: float

        Mean.

    std: float

        Standard deviation.

    pfail: float

        Probability that the mark will be 0.

    pnan: float

        Probability that the mark will be ``np.nan``.

    Returns
    -------
    pd.Series

        An assignment.

    Examples
    --------
    >>> import fakeitmakeit as fm
    ...
    >>> fm.assignment(["abc123", "xyz321"])  # doctest: +SKIP
    username
    abc123     67.9
    xyz321    73.08
    Name: mark, dtype: Float64

    """
    usernames = list(usernames)
    invalid = [u for u in usernames if not fmiv.username(u)]
    if invalid:
        raise ValueError(f"Invalid usernames: {invalid}.")

    return pd.Series(
        data=[mark(mean=mean, std=std, pfail=pfail, pnan=pnan) for _ in usernames],
        index=pd.Index(usernames, name="username"),
        name="mark",
        dtype=np.float64,  # allow missing values
    )
